package sysExt ;

/*
 * @(#)methodRef.java	1.0 96/01/05  Sal Cataudella
 *
 * Copyright (c) 1995 Sal Cataudella. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please contact author for other arrangements.
 *
 */

import java.io.* ;


import java.util.Hashtable ;

/** 
* Abstraction of a method reference.
* Used for dynamic method dispatching, aka a method callback mechanism.
* Several convenience methods are provided.
* Examples:
* <pre>
*  import testPkg.* ; // contains greeter class..
*  import sysExt.* ;  // contains methodRef class
*  
*  // Create a target object..
*  greeter aGreeter = new greeter (1);
*
*  // 'point' to sayHello() action, and set the target to aGreeter.
*  methodRef mr = methodRef.newRef(aGreeter, "sayHello"); 
*
*  // make aGreeter 'perform' sayHello()
*  mr.doAction();
*         .
*         .
*  // now point to sayGoodBye() action..
*  methodRef mr2 = methodRef.newRef("testPkg",
*                                   "greeter",
*                                   "sayGoodbye");
*
*  // make aGreeter 'perform' sayGoodBye()
*  mr2.on(aGreeter).doAction();
* </pre>
*	
* A unique subclass of methodRef is generated as the applet/application runs. 
* The user of the methodRef object however,  *always* deals with it as
* a methodRef. The dynamically created subclass has the name
* CLASS__METHOD__PARAMCLASS
* where CLASS is the fully qualified name of the method's class,
* METHOD is the method's name, and PARAMCLASS is the fully qualified method's
* parameters' class name. This subclass is made part of a methodRefClasses
* package.
* 
* <pre>
* Restrictions:
	- IMPORTANT: Because the generated methodRef lives in its own package,
	only PUBLIC methods and PUBLIC interfaces of PUBLIC classes can be
	pointed to. Likewise, the target method's parameter class must be
	PUBLIC as well. 
	   
	- Once a method reference is generated, it is bound to the
	referred method and cannot be changed.

	- Applets may not use this package. This is because of a current security
	restriction on class loaders.
* </pre>
*
* Known bugs:
* - Specifying an invalid method name to methodRef.newRef(), does not
* always cause a NoSuchMethodError exception to be generated in doAction().
* Since it sometimes occurs, I assume it's a bug in the VM.
* This is presently being looked at...
*
* @author		Sal Cataudella
* @version        1.00     Jan 7 1996
* 
*
*
*/

public abstract class methodRef extends Object {
    /**
    * The target of the callback method
    *
    */
    protected Object O = null ;

    /**
    * Constructor
    */
    public methodRef(){};

    /**
    * Performs the action on the 
    * on the target object..This is an abstract
    * method that is overridden in the dynamically
    * generated method reference (functor)
    * Invocation of doAction() will cause
    * your desired method to be called and
    * passed o as a parameter..
    * @param o	passed to called back method	
    */ 
    public abstract void doAction( Object o ) ; 


    /** 
    * Sets the target object
    * of the action
    * @param target	Target of referenced method
    */
    public final void setTarget ( Object target )
    {
    	O = target ;
    } ;




    /** 
    * Gets the target object
    * of the action
    * @return	Current target of referenced method.
    */
    public final Object getTarget ()
    {
    	return O;
    } ;



    /**
    * overloads default doAction() 
    * (convenience function) - Passes a null reference to the referred method..
    * Use this one if parameter to method callback need not be used.
    */
    public void doAction()  { 
	doAction(null) ;
    }

    /**
    * Convenience method.
    * Useful for dispatching a method on
    * an target object all in one statement...
    *
    * <pre>
    * usage: mr.on( obj ).doAction() ;
    * </pre>
    *
    * @param o       target object of action.
    *                Note: The target is set to this object until changed.
    *
    * @return	   reference to self (i.e. this
    *                subclass of a methodRef)
    */
    public methodRef on( Object o ) {
	setTarget(o) ;
	return this ;
    } ;



    // Does the actual class generation...
    private static ClassBuilder aClassBuilder = new ClassBuilder() ;

    // should protect access of method be overidden
    private static boolean overrideProtectStatus = false ;

    /**
    * Creates a reference to a specific method.
    * The method *must* have a specific signature, i.e
    * must take a parameter that derives from Object and returns void.
    * In order to 'point' a specific method, a subclass of
    * methodRef is created at *runtime*, with Java byte
    * codes dynamically generated to invoke the method
    * within doAction().
    *
    * @param packageString		package of target object
    * @param classString		class of target object
    * @param methodString		string representation of method
    *						to make a reference to  ..
    * @param paramClassString	string representation of method's parameter's class
    * @return				reference to method
    * @exception			InvalidMethodRefException If generation of methodRef fails for any reason.
    */	
    public
    static methodRef newRef(
                            	String packageString,
					String classString,
                            	String methodString,
					String paramClassString )
						throws InvalidMethodRefException {
     try {
	methodRef 	mr = null ;


	String paramClass = new String ( paramClassString.replace( '.', '_' ) );	
  
      String functorClassName =	(packageString != null ?
							packageString.replace('.', '_' ):
							"_")
						+ "_" +
						classString + "_" +
						methodString + "_" + paramClass;
	if( packageString != null )
		packageString = packageString.replace( '.', '/' ) ;

      String  functorClassSignature =
                  "methodRefClasses." //remember to make this settable...
			+ functorClassName ;

	Class s ;

	if( !useCompiler )
      //
      // Build the class dynamically
      //
	s = aClassBuilder.buildClassFile(
						packageString,
						classString,
						methodString,
						functorClassSignature,
						functorClassName,
						paramClassString,
						methodRef.overrideProtectStatus );

	else try {
		s = aClassBuilder.useCompiler(
						packageString,
						classString,
						methodString,
						functorClassSignature,
						functorClassName,
						paramClassString,
						methodRef.overrideProtectStatus  );
	} catch( Exception e ) {
		throw new ClassNotFoundException() ;
	}
	//
	// Once class has been generated, instantiate
	// an object from that class...
	//
	mr = (methodRef)s.newInstance();
	return mr;	
     } catch( Exception e ) {
		throw new InvalidMethodRefException(
			"*** newRef( String, String, String, String ) >> "
			+ e.toString() + " << " ) ;
     }


    }


 

    /**
    * Creates a reference to a specific method.Convenience method.
    * Tthe target method's  parameter's class defaults to Object...
    * @param packageString		package of target object
    * @param classString		class of target object
    * @param methodString		string representation of method
    *					to make a reference to  ..
    * @return				reference to method
    * @exception			InvalidMethodRefException If generation of methodRef fails for any reason  
    */	
    public
    static methodRef newRef(
				String packageString,
				String classString,
				String methodString)
					throws InvalidMethodRefException {


	return newRef( packageString, classString,methodString,"java.lang.Object" ) ;

    }

   /**
    * newRef() - Convenience method. No need to specify
    * a package string and class string. All of that is extracted 
    * from the targetObj. Note: targetObj must have been
    * instantiated first.The target O is then set to targetObj.
    * The method's parameter's class defaults to Object... 
    *
    * The target object is then set to targetObj.
    * @param targetObj	package and clas string are extracted from
    *				it. The target object is then set to this.
    * @param method		string representation of method to make
    *			       reference to. 
    * @return			reference to method
    * @exception		InvalidMethodRefException If generation of methodRef fails for any reason
    */
    public static methodRef newRef( Object targetObj, String method )
						throws InvalidMethodRefException {


		
	// Extract the package and class info
	// from the targetObj..
	//
	String s = 	targetObj.getClass().getName() ;
	String classString ;
	String packageString;
	int index = s.lastIndexOf( '.' ) ;

	classString = s.substring( index + 1, s.length() ) ;
	if( index != -1 ) {
		packageString = s.substring( 0, index ) ;
	} else
		packageString = null ;


	//
	// Create a methodRef and set its target to targetObj
	//
	methodRef mr = newRef( packageString, classString, method, "java.lang.Object" ) ;
	mr.setTarget( targetObj ) ;

	return mr ;
    }

 


    /**
    * newRef() - Convenience method. No need to specify
    * a package string and class string. All of that is extracted 
    * from the targetObj. Note: targetObj must have been
    * instantiated first.
    * The target object is then set to targetObj.
    * @param targetObj	package and clas string are extracted from
    *				it. The target object is then set to this.
    * @param method	string representation of target method to make
    *			       reference to. 
    * @param paramClass  Class of target method's parameter
    * @return		reference to method
    * @exception		InvalidMethodRefException If generation of methodRef fails for any reason
    */
    public static methodRef newRef( Object targetObj, String method, Class paramClass)
						throws InvalidMethodRefException {


		
	// Extract the package and class info
	// from the targetObj..
	//

	String s = 	targetObj.getClass().getName() ;
	String classString ;
	String packageString;
	int index = s.lastIndexOf( '.' ) ;

	classString = s.substring( index + 1, s.length() ) ;
	if( index != -1 ) {
		packageString = s.substring( 0, index ) ;
	} else
		packageString = null ;

	//
	// Create a methodRef and set its target to targetObj
	//
	methodRef mr = newRef( packageString, classString, method, paramClass.getName()  ) ;
	mr.setTarget( targetObj ) ;

	return mr ;
    }

   /**
    * flag to determine if javac compiler should be used to generate
    * functor class. As this is exceptionally slow as well as
    * extremely applet unsafe, this feature is left here for debugging
    * purposes only. 
    */
    static public boolean useCompiler = false ;


   /**
    * Controls whether a .class file should be written to disk.
    * Requires a ./methodRefClasses subdirectory to be present
    * for the resulting class files. Default is not to write to disk.
    */
    public static void toDisk( boolean bool ) {
	ClassBuilder.debug = bool ;
    }
    /**
    * Allows you to 'point' to a protected method...
    * Not yet implemented.
    */
    public static void overrideProtect( boolean val) {
		overrideProtectStatus = val;
    }

}

